from typing import List
from importlib import resources
import logging

logger = logging.getLogger(__name__)

DATA_PATH = resources.files(f"rara_text_evaluator.data")
SUPPORTED_BUILD_LANGUAGES_FILE_PATH = DATA_PATH / "supported_build_languages.txt"

def load_languages(file_path: str) -> List[str]:
    with open(file_path, "r") as f:
        languages = [
            lang.strip()
            for lang in f.read().split("\n")
        ]
    return languages

def get_supported_build_languages(file_path: str = SUPPORTED_BUILD_LANGUAGES_FILE_PATH) -> List[str]:
    try:
        # Try loading supported languages straight from langdetect
        from langdetect.detector_factory import DetectorFactory, PROFILES_DIRECTORY

        factory = DetectorFactory()
        factory.load_profile(PROFILES_DIRECTORY)

        supported_build_languages = factory.langlist

    except Exception as e:
        # Use hardcoded version in case something has changed in langdetect
        # and some functions are no longer working as expected
        log_message = (
            f"Detecting supported build languages with `langdetect` failed with exception: '{e}'. "
            "Trying to load languages from file instead (the list might not be up to date)."
        )
        logger.debug(log_message)
        supported_build_languages = load_languages(file_path)

    logger.debug(f"Finished loading {len(supported_build_languages)} supported build languages.")
    return supported_build_languages



class InvalidInputError(Exception):
	"""Raised when something incorrect given to parse."""
	pass

class InvalidRegistryError(Exception):
    """Raised when fallback key is missing from the model registry."""
    pass



import pathlib

import pytest
import warnings

from rara_text_evaluator.ngram_model_builder import NgramModelBuilder
from rara_text_evaluator.quality_evaluator import QualityEvaluator, DEFAULT_RESPONSE

DEFAULT_FILENAME = "ngram_model.pkl"


@pytest.fixture(autouse=True)
def delete_model_file(request):
    if "override" in request.keywords:
        yield
    else:
        yield
        path = pathlib.Path(DEFAULT_FILENAME)
        path.unlink()
        assert path.exists() is False


def test_model_training_process_with_et_text(delete_model_file):
    language = "et"

    qv = QualityEvaluator()
    test_text_en = "You should be aware that asking whether an african or european swallow is faster is inappropriate!"
    test_text_et = "Ta lisas, et Eesti meretuuleparke alles planeeritakse ning et sinna, kus need hakkavad..."

    en_probability_default_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_default_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")


    nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="abcdefghijklmnopqrstuvwxyzõäöü.,:;-_!\"()%@1234567890' ")
    text_corpus = "Kas sa oled teadlik sellest et märgade mõõkade loopimine ei ole alus ühe õiglase valitsuse loomiseks!?"
    nmb.build_model(text_corpus)
    nmb.save_model(DEFAULT_FILENAME)

    assert pathlib.Path(DEFAULT_FILENAME).exists()


    qv.add_model(lang=language, model_path=DEFAULT_FILENAME)
    en_probability_new_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_new_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")

    # The probability for EN text should not change as the model remains the same
    assert en_probability_default_model == en_probability_new_model

    # The probability for ET text should change as we replaced the default model
    # with the new model we trained
    assert et_probability_default_model != et_probability_new_model


def test_model_training_process_with_en_text(delete_model_file):
    language = "en"
    qv = QualityEvaluator()
    test_text_en = "You should be aware that asking whether an african or european swallow is faster is inappropriate!"
    test_text_et = "Ta lisas, et Eesti meretuuleparke alles planeeritakse ning et sinna, kus need hakkavad..."

    en_probability_default_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_default_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")


    nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="abcdefghijklmnopqrstuvwxyzõäöü.,:;-_!\"()%@1234567890' ")
    text_corpus = "Kas sa oled teadlik sellest et märgade mõõkade loopimine ei ole alus ühe õiglase valitsuse loomiseks!?"
    nmb.build_model(text_corpus)
    nmb.save_model(DEFAULT_FILENAME)

    assert pathlib.Path(DEFAULT_FILENAME).exists()


    qv.add_model(lang=language, model_path=DEFAULT_FILENAME)
    en_probability_new_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_new_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")

    # The probability for ET text should not change as the model remains the same
    assert et_probability_default_model == et_probability_new_model

    # The probability for EN text should change as we replaced the default model
    # with the new model we trained
    assert en_probability_default_model != en_probability_new_model


def test_model_training_process_with_et_iterator(delete_model_file):
    language = "et"

    qv = QualityEvaluator()
    test_text_en = "You should be aware that asking whether an african or european swallow is faster is inappropriate!"
    test_text_et = "Ta lisas, et Eesti meretuuleparke alles planeeritakse ning et sinna, kus need hakkavad..."

    en_probability_default_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_default_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")


    nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="abcdefghijklmnopqrstuvwxyzõäöü.,:;-_!\"()%@1234567890' ")
    text_corpus = [
        "Kas sa oled",
        "teadlik sellest et",
        "märgade mõõkade loopimine",
        "ei ole alus ühe õiglase valitsuse loomiseks!?"
    ]
    text_iterator = (text for text in text_corpus)
    nmb.build_model(text_corpus)
    nmb.save_model(DEFAULT_FILENAME)

    assert pathlib.Path(DEFAULT_FILENAME).exists()


    qv.add_model(lang=language, model_path=DEFAULT_FILENAME)
    en_probability_new_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_new_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")

    # The probability for EN text should not change as the model remains the same
    assert en_probability_default_model == en_probability_new_model

    # The probability for ET text should change as we replaced the default model
    # with the new model we trained
    assert et_probability_default_model != et_probability_new_model
    assert et_probability_new_model > 0


def test_model_training_process_with_en_iterator(delete_model_file):
    language = "en"
    qv = QualityEvaluator()
    test_text_en = "You should be aware that asking whether an african or european swallow is faster is inappropriate!"
    test_text_et = "Ta lisas, et Eesti meretuuleparke alles planeeritakse ning et sinna, kus need hakkavad..."

    en_probability_default_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_default_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")


    nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="abcdefghijklmnopqrstuvwxyzõäöü.,:;-_!\"()%@1234567890' ")
    text_corpus = [
        "You should be aware",
        "that asking whether",
        "an african or european swallow",
        "is inappropriate!?"
    ]
    text_iterator = (text for text in text_corpus)
    nmb.build_model(text_corpus)
    nmb.save_model(DEFAULT_FILENAME)

    assert pathlib.Path(DEFAULT_FILENAME).exists()


    qv.add_model(lang=language, model_path=DEFAULT_FILENAME)
    en_probability_new_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_new_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")

    # The probability for ET text should not change as the model remains the same
    assert et_probability_default_model == et_probability_new_model

    # The probability for EN text should change as we replaced the default model
    # with the new model we trained
    assert en_probability_default_model != en_probability_new_model
    assert en_probability_new_model > 0

def test_model_training_process_with_ru_text(delete_model_file):
    language = "ru"

    qv = QualityEvaluator()
    test_text_en = "You should be aware that asking whether an african or european swallow is faster is inappropriate!"
    test_text_et = "Ta lisas, et Eesti meretuuleparke alles planeeritakse ning et sinna, kus need hakkavad..."
    test_text_ru = "От планов до строительства еще нужно пройти долгий путь, на этапе планирования у нас"

    en_probability_default_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_default_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")
    ru_probability_default_model = qv.get_probability(text=test_text_ru, length_limit=1, lang="ru")


    nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="бфждёпнчргейъзлвтьэсхщюуишмыкоаця.,:;-_!\"()%@1234567890' ")
    text_corpus = "Мы уже высказали свою точку зрения и будем продолжать это делать. Это требует очень точных!?"
    nmb.build_model(text_corpus)
    nmb.save_model(DEFAULT_FILENAME)

    assert pathlib.Path(DEFAULT_FILENAME).exists()

    qv.add_model(lang=language, model_path=DEFAULT_FILENAME)
    en_probability_new_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_new_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")
    ru_probability_new_model = qv.get_probability(text=test_text_ru, length_limit=1, lang="ru")

    # The probability for EN text should not change as the model remains the same
    assert en_probability_default_model == en_probability_new_model

    # The probability for ET text should not change as the model remains the same
    assert et_probability_default_model == et_probability_new_model

    # The probability for RU text should change as we replaced the default model
    # with the new model we trained
    assert ru_probability_default_model != ru_probability_new_model
    assert ru_probability_new_model > 0


def test_model_training_process_with_custom_lang_text(delete_model_file):
    language = "klingon"

    qv = QualityEvaluator()
    test_text_en = "You should be aware that asking whether an african or european swallow is faster is inappropriate!"
    test_text_et = "Ta lisas, et Eesti meretuuleparke alles planeeritakse ning et sinna, kus need hakkavad..."
    test_text_kl = "vaj nItebHa'vam vIghoS,"

    en_probability_default_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_default_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")
    kl_probability_default_model = qv.get_probability(text=test_text_kl, length_limit=1)

    with warnings.catch_warnings(record=True) as warning_record:
        nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="jilnrguhqtmpseybvwaocd.,:;-_!\"()%@1234567890' ")
        text_corpus = "'ach cha'logh nItebHa' 'ej mIw vIghoS 'e' vIghoS, 'each vIghoS 'ej cha'logh vIghoS"
        nmb.build_model(text_corpus)
        nmb.save_model(DEFAULT_FILENAME)

    # A warning should be thrown for a language not supported by langdetect
    print(warning_record[-1])
    assert len(warning_record) == 1
    assert warning_record[0].category == UserWarning

    assert pathlib.Path(DEFAULT_FILENAME).exists()

    qv.add_model(lang=language, model_path=DEFAULT_FILENAME)
    en_probability_new_model = qv.get_probability(text=test_text_en, length_limit=1, lang="en")
    et_probability_new_model = qv.get_probability(text=test_text_et, length_limit=1, lang="et")
    kl_probability_new_model = qv.get_probability(text=test_text_kl, length_limit=1, lang="klingon")

    # The probability for EN text should not change as the model remains the same
    assert en_probability_default_model == en_probability_new_model

    # The probability for ET text should not change as the model remains the same
    assert et_probability_default_model == et_probability_new_model

    # The probability for klingon text should change as we replaced the default model
    # with the new model we trained
    assert kl_probability_default_model != kl_probability_new_model
    assert kl_probability_new_model > 0

@pytest.mark.override
def test_warning_is_thrown_for_unsupported_language():
    language = "bleh"
    with warnings.catch_warnings(record=True) as warning_record:
        nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="abcde")

    assert len(warning_record) == 1
    assert warning_record[0].category == UserWarning

@pytest.mark.override
def test_warning_is_not_thrown_for_supported_language():
    language = "ru"
    with warnings.catch_warnings(record=True) as warning_record:
        nmb = NgramModelBuilder(n=2, lang=language, accepted_chars="abcde")

    assert len(warning_record) == 0



import pytest

from rara_text_evaluator import exceptions
from rara_text_evaluator.quality_evaluator import QualityEvaluator, DEFAULT_RESPONSE, DEFAULT_TEXT_SIZE_REQUIREMENT


def test_quality_evaluator_process():
    evaluator = QualityEvaluator()
    text = "Elas kord üks muinasjutt ja see muinasjutt sai läbi!!"
    probability = evaluator.get_probability(text)
    assert len(text) > DEFAULT_TEXT_SIZE_REQUIREMENT
    assert probability > DEFAULT_RESPONSE


def test_short_text_returning_default_response():
    evaluator = QualityEvaluator()
    text = "See on lühike!"
    probability = evaluator.get_probability(text)
    assert len(text) < DEFAULT_TEXT_SIZE_REQUIREMENT
    assert probability == DEFAULT_RESPONSE


def test_fallback_language_being_used():
    evaluator = QualityEvaluator()
    text = ("Maanantaista torstaihin herään viisitoista yli seitsemän. "
            "Perjantaisin herään kahdeksalta, koska oppituntini alkavat myöhemmin. "
            "Herättyäni syön aamupalaa, puen päälleni ja menen kouluun.!")
    evaluator.get_probability(text)
    assert evaluator.model_lang is None


def test_using_defined_language_instead_of_autodetect():
    evaluator = QualityEvaluator()
    text = "Milline on kiirem, euroopa või aafrika pääsuke? Ma tahan teada!"
    evaluator.get_probability(text, lang="en")
    assert evaluator.model_lang == "en"


def test_using_wrong_language_code_results_in_fallback():
    evaluator = QualityEvaluator()
    text = "Milline on kiirem, euroopa või aafrika pääsuke? Ma tahan teada!"
    language_code = "fin"
    evaluator.get_probability(text, lang=language_code)
    assert language_code not in evaluator.language_model_container
    assert evaluator.model_lang is None


def test_different_texts_returning_unique_test_quality_answers():
    evaluator = QualityEvaluator()
    texts = [
        "Teadavasti, kui inimene on kergem kui part, siis ta on nõid!",
        "Võtke pühalt käsigranaadilt tema kaitse ja lugege arve kolm, ei loe mitte kaks ja kindlasti mitte neli.",
        "Mis on su nimi? Kust sa tuled? Mis on ühe pääsukese maksimum lendamiskiirus?"
    ]

    qualities = []
    for text in texts:
        qualities.append(evaluator.get_probability(text))

    assert len(set(qualities)) == len(texts)


def test_exception_being_thrown_when_no_models_loaded():
    evaluator = QualityEvaluator(load_defaults=False)
    text = "Milline on kiirem, euroopa või aafrika pääsuke? Ma tahan teada!"
    with pytest.raises(exceptions.InvalidRegistryError):
        evaluator.get_probability(text, lang="et")


def test_updating_length_limit():
    evaluator = QualityEvaluator()
    test_text = "tere tere vana kere"

    probability_1 = evaluator.get_probability(test_text, lang="et")
    assert probability_1 == DEFAULT_RESPONSE

    probability_2 = evaluator.get_probability(test_text, lang="et", length_limit=5)
    assert probability_2 != DEFAULT_RESPONSE


def test_updating_default_response():
    evaluator = QualityEvaluator()
    test_text = "tere tere vana kere"
    length_limit = 50
    language = "et"

    probability_1 = evaluator.get_probability(
        test_text,
        lang=language,
        length_limit=length_limit
    )
    assert probability_1 == DEFAULT_RESPONSE

    updated_default_response = -1

    probability_2 = evaluator.get_probability(
        test_text,
        lang=language,
        length_limit=length_limit,
        default_response=updated_default_response
    )
    assert probability_2 != DEFAULT_RESPONSE
    assert probability_2 == updated_default_response


def test_is_valid_et():
    test_text = "Wanemuine. Piihapüewal, 7. weebruaril s. a, Kmemllk KM. Operett 3 järgus. Planquette muusika. Kakatus kess 8 õhtu."
    evaluator = QualityEvaluator()
    evaluator.set_threshold(lang="et", threshold=0.8)

    is_valid_1 = evaluator.is_valid(test_text, lang="et")
    assert not is_valid_1

    evaluator.set_threshold(lang="et", threshold=0.4)
    is_valid_2 = evaluator.is_valid(test_text, lang="et")
    assert is_valid_2


def test_restoring_default_thresholds():
    evaluator = QualityEvaluator()
    default_threshold = evaluator.default_threshold

    et_old_threshold = evaluator.thresholds["et"]
    et_new_threshold = 0.234
    assert et_old_threshold != et_new_threshold
    assert default_threshold == et_old_threshold

    evaluator.set_threshold(lang="et", threshold=et_new_threshold)
    assert et_new_threshold == evaluator.thresholds["et"]
    assert default_threshold != evaluator.thresholds["et"]

    evaluator.set_threshold(lang="et", threshold=evaluator.default_threshold)
    assert evaluator.thresholds["et"] == default_threshold


def test_invalid_threshold_setting():
    evaluator = QualityEvaluator()
    # Check that erroris raised with invalid language
    with pytest.raises(exceptions.InvalidInputError) as exception_info:
        evaluator.set_threshold(lang="blablabla", threshold=0.8)

    # Check that error is raised with invalid threshold
    with pytest.raises(exceptions.InvalidInputError) as exception_info:
        evaluator.set_threshold(lang="et", threshold=7.6)

    # Check that NO errors are raise with valid input
    evaluator.set_threshold(lang="et", threshold=0.6)



import logging
import math
import pickle
from importlib import resources
from typing import NoReturn, Tuple, Dict
from langdetect import detect, DetectorFactory, LangDetectException
from . import exceptions

logger = logging.getLogger(__name__)

DEFAULT_RESPONSE = 0.0
DEFAULT_TEXT_SIZE_REQUIREMENT = 30


class QualityEvaluator:
    FALLBACK_MODEL_KEY = "fallback"

    def __init__(self,
            load_defaults: bool = True,
            default_threshold: float = 0.7
        ):
        """
        Initializes the QualityValidator by loading the primary and optional fallback models.

        Parameters
        ----------
        load_defaults : bool
            Whether to automatically load the default models into the memory.
            If set to False, models including a fallback one must be loaded using the add_model() method.

        default_threshold: float
            Default threshold for distinguishing valid texts from invalid ones.
            This default is applied to all languages, but can be modified with `set_threshold`.
        """
        DetectorFactory.seed = 0  # Set seed for langdetect to ensure consistent results

        self.default_threshold: float  = default_threshold

        self.default_model_mapping: dict = {
            "et": "text_validator_ngram_3_et.pkl",
            "en": "text_validator_ngram_3_en.pkl",
            self.FALLBACK_MODEL_KEY: "text_validator_ngram_3_fallback.pkl",
        }

        self.thresholds = dict = {
            lang: self.default_threshold
            for lang in self.default_model_mapping.keys()
        }

        self.language_model_container: dict = {}
        if load_defaults:
            self._load_default_language_models()

    def _load_default_language_models(self) -> None:
        logger.debug(f"Loading default language models...")
        for language, name in self.default_model_mapping.items():
            path = resources.files(f"rara_text_evaluator.models")
            self.language_model_container[language] = self._load_model(path / name)
            logger.debug(f"Loading model for '{language}': {path}/{name}...")

    def _load_model(self, model_path: str):
        with open(model_path, "rb") as f:
            return pickle.load(f)

    def set_threshold(self, lang: str, threshold: float) -> NoReturn:
        """
        Set threshold for a specific language. Function `is_valid`
        will use this threshold to determine whether a given text
        is valid or not.

        Parameters
        ----------
        lang : str
            Language to apply the threshold for.
        threshold: float [0,1]
            Threshold in range [0,1] for separating valid texts from invalid ones.
        Returns
        -------
        None
        """

        if lang not in self.language_model_container:
            supported_langs = list(self.language_model_container.keys())
            error_message = (
                f"No model for language '{lang}'. "
                f"Currently supports models for languages: {supported_langs}."
                "Use `NgramModelBuilder` to train a new model."
            )
            raise exceptions.InvalidInputError(error_message)

        if not (0 <= threshold <=1):
            error_message = (
                f"Invalid value for parameter `threshold` ({threshold}). "
                "Threshold should be a float in range [0,1]."
            )
            raise exceptions.InvalidInputError(error_message)

        self.thresholds[lang] = threshold
        logger.debug(f"Default threshold for language '{lang}' is set to {threshold}.")

    def _set_model(self, model_data: dict) -> NoReturn:
        """
        Set the current model to use based on the model data dictionary.

        Parameters
        ----------
        model_data : dict
            The model data dictionary containing model parameters and data.

        Returns
        -------
        None
        """
        self.model_data = model_data
        self.n = model_data.get("n_gram")
        self.model_lang = model_data.get("lang", None)
        self.accepted_chars = model_data.get("accepted_chars")
        self.log_probabilities = model_data.get("log_probabilities")
        self.char_count = len(self.accepted_chars)

    def _normalize(self, line: str) -> str:
        """
        Normalize text by converting to lowercase and removing unaccepted characters.
        """
        return "".join([c.lower() for c in line if c.lower() in self.accepted_chars])

    def _ngram(self, n: int, text: str) -> Tuple[str, str]:
        """
        Return all n-grams from text after normalizing.
        """
        filtered = self._normalize(text)
        for start in range(len(filtered) - n + 1):
            ngram = filtered[start:start + n]
            prefix = ngram[:-1]
            suffix = ngram[-1]
            yield (prefix, suffix)

    def _get_avg_transition_prob(self, text: str) -> float:
        """
        Returns the average transition probability from the text using log probabilities.
        """
        log_prob = 0.0
        transition_ct = 0

        for prefix, suffix in self._ngram(self.n, text):
            if prefix in self.log_probabilities and suffix in self.log_probabilities[prefix]:
                log_prob += self.log_probabilities[prefix][suffix]
            else:
                log_prob += math.log(1 / (self.char_count * 2))  # Smoothing for unseen n-grams
            transition_ct += 1

        return math.exp(log_prob / (transition_ct or 1))

    def _detect_language(self, text: str) -> str | None:
        """
        Detect the language of the provided text, return None if detection fails
        (for example if the text contains no alphabetic characters).

        Parameters
        ----------
        text : str
            Input text to detect language from.

        Returns
        -------
        str
            Detected language code (ISO-639-1) or None if detection fails.
        """

        try:
            return detect(text)
        except LangDetectException:
            return None

    def _detect_and_set_model(self, text: str, lang: str | None = None) -> NoReturn:
        """
        Validate the detected language and switch to the appropriate model if necessary.
        Skips validation if no fallback model is provided.

        Parameters
        ----------
        text : str
            The text for which the language needs to be validated.

        lang : str
            Language code for the given text, if set to None will use langdetect to automatically detect the language.

        Returns
        -------
        None
        """
        detected_lang = lang or self._detect_language(text)
        logger.debug(f"Using language '{detected_lang}' for text '{text[:20]}...'")

        if detected_lang in self.language_model_container:
            self._set_model(self.language_model_container[detected_lang])
            logger.debug(f"{detected_lang} == {self.model_lang}, using primary model.")
        else:
            self._set_model(self.language_model_container[self.FALLBACK_MODEL_KEY])
            logger.debug(f"{detected_lang} != {self.model_lang}, using fallback model.")

    def _sigmoid(self, x: float, x0: float = 0.05, k: float = 50.0) -> float:
        """
        Sigmoid function to calculate the probability based on the average transition probability.

        Parameters
        ----------
        x : float
            The average transition probability.
        x0 : float, optional
            The x value of the sigmoid midpoint, by default 0.05.
        k : float, optional
            The steepness of the sigmoid, by default 50.0.

        Returns
        -------
        float
            The probability that the text is readable.
        """
        return 1 / (1 + math.exp(-k * (x - x0)))

    def _validate_model_registry(self):
        if self.FALLBACK_MODEL_KEY not in self.language_model_container:
            raise exceptions.InvalidRegistryError("There must always be a fallback model loaded into memory!")


    def add_model(self, lang: str, model_path: str) -> None:
        """
        Allows the user to add a model from a file path into the model storage,
        either to overwrite existing languages or adding new ones.

        To overwrite the fallback model use the QueryValidator.FALLBACK_MODEL_KEY
        as the language parameter.

        Parameters
        ----------
        lang : str
            Language parameter of the new model.
        model_path: str
            Path to load the model from.
        """

        logger.debug(f"Adding a new model for language '{lang}' from path '{model_path}'...")
        self.language_model_container[lang] = self._load_model(model_path)


    def get_probability(
            self,
            text: str,
            default_response: float = DEFAULT_RESPONSE,
            length_limit: int = DEFAULT_TEXT_SIZE_REQUIREMENT,
            lang: str | None = None
        ) -> float:
        """
        Calculate the probability that the given text seems readable and is not full of artifacts.
        Method uses a sigmoid function to calculate the probability based on the average transition probability.

        Parameters
        ----------
        text : str
            The text to calculate the probability for.

        default_response : float
            What number to return if the text doesn't reach the required length.

        length_limit : int
            Maximum number of characters for text without leading or trailing whitespace.

        lang : str
            Language code for the given text, if set to None will use langdetect to automatically detect the language.

        Returns
        -------
        float
            The probability that the text is readable
        """
        self._validate_model_registry()

        if len(text.strip()) < length_limit:
            log_message = (
                f"Text must be >={length_limit} characters long without leading or trailing whitespace. "
                f"Returning default response: {default_response}."
            )
            logger.debug(log_message)
            return default_response

        self._detect_and_set_model(text, lang=lang)
        sigmoid_prob = self._sigmoid(self._get_avg_transition_prob(text))

        return sigmoid_prob

    def is_valid(self,
            text: str,
            default_response: float = DEFAULT_RESPONSE,
            length_limit: int = DEFAULT_TEXT_SIZE_REQUIREMENT,
            lang: str | None = None
        ) -> bool:
        """
        Determines whether an input text is valid or not.

        Parameters
        ----------
        text : str
            The text to calculate the probability for.

        default_response : float
            What number to return if the text doesn't reach the required length.

        length_limit : int
            Maximum number of characters for text without leading or trailing whitespace.

        lang : str
            Language code for the given text, if set to None will use langdetect to automatically detect the language.

        Returns
        -------
        bool
            Boolean value indicating whether the input text has sufficient quality or not.

        """
        detected_lang = lang or self._detect_language(text)

        probability = self.get_probability(
            text=text,
            default_response=default_response,
            length_limit=length_limit,
            lang=detected_lang
        )
        if detected_lang not in self.thresholds:
            detected_lang = self.FALLBACK_MODEL_KEY

        if probability >= self.thresholds[detected_lang]:
            return True

        return False



import math
import pickle
import warnings
import logging
import os

from collections import defaultdict
from pathlib import Path
from typing import Iterator, Union, Dict, NoReturn, List

from . import utils


logger = logging.getLogger(__name__)


LANGDETECT_SUPPORTED_LANGS = utils.get_supported_build_languages()

class NgramModelBuilder:
    def __init__(
            self,
            n: int = 2,
            lang: str | None = None,
            accepted_chars: str = "abcdefghijklmnopqrstuvwxyzõäöü.,:;-_!\"()%@1234567890' "
    ):
        """
        Initialize the NgramModelBuilder.

        Parameters
        ----------
        n : int
            The size of the n-grams (e.g., 2 for bigrams, 3 for trigrams).
        lang : str
            Language code (ISO-639-1).
        accepted_chars : str
            A string of accepted characters for the n-gram model.

        Warnings
        ------
        UserWarning
            If the lang parameter is not a valid ISO-639-1 code supported by `langdetect`.
        """

        if lang is not None and lang not in LANGDETECT_SUPPORTED_LANGS:
            warning_message = (
                f"Language '{lang}' is not supported by `langdetect`. "
                "Are you sure you want to use it? "
                "`langdetect` supports language codes in ISO-639-1 format. "
                f"Complete list of languages supported by `langdetect`: {LANGDETECT_SUPPORTED_LANGS}."
            )
            warnings.warn(warning_message)
            logger.warning(f"Initiating a NgramModelBuilder object with language not supported by `langdetect` ('{lang}').")

        self.n = n
        self.lang = lang
        self.accepted_chars = accepted_chars
        self.char_count = len(self.accepted_chars)
        self.ngram_counts = defaultdict(lambda: defaultdict(int))
        self.log_probabilities = defaultdict(dict)  # Prefix -> {Suffix: log probability}

    def _normalize(self, text: str) -> str:
        """
        Normalize text by converting to lowercase and removing unaccepted characters.

        Parameters
        ----------
        text : str
            Text to be normalized.

        Returns
        -------
        str
            Normalized text.
        """
        return "".join([c.lower() for c in text if c.lower() in self.accepted_chars])

    def build_model(self, text_corpus: str | Iterator[str]) -> NoReturn:
        """
        Build the n-gram model based on the given text corpus.

        Parameters
        ----------
        text_corpus : str | Iterator[str]
            Text or iterator to build the n-gram model from.

        Returns
        -------
        None
        """
        logger.debug(f"Start training a model for language '{self.lang}'...")

        # If string is passes as a text_corpus,
        # reformat it as a list to make it compatible
        # with iterator output
        if isinstance(text_corpus, str):
            text_corpus = [text_corpus]

        for text in text_corpus:

            normalized_text = self._normalize(text)

            for i in range(len(normalized_text) - self.n + 1):
                ngram = normalized_text[i:i + self.n]
                prefix = ngram[:-1]
                suffix = ngram[-1]
                self.ngram_counts[prefix][suffix] += 1

        for prefix, suffix_counts in self.ngram_counts.items():
            total_count = sum(suffix_counts.values())
            for suffix in self.accepted_chars:
                count = self.ngram_counts[prefix].get(suffix, 0) + 1  # Add-one smoothing
                prob = count / total_count
                self.log_probabilities[prefix][suffix] = math.log(prob)

        logger.debug(f"Finished training a text quality evaluation model for language '{self.lang}'!")


    def save_model(self, file_path: str) -> NoReturn:
        """
        Save the model as a dictionary to a pickle file.

        Parameters
        ----------
        file_path : str
            The model save path.

        Returns
        -------
        None
        """

        Path(file_path).parent.mkdir(parents=True, exist_ok=True)

        model_dict = {
            "n_gram": self.n,
            "accepted_chars": self.accepted_chars,
            "log_probabilities": self.log_probabilities
        }

        if self.lang:
            model_dict["lang"] = self.lang

        with open(file_path, "wb") as f:
            pickle.dump(model_dict, f)

        logger.debug(f"Saved text quality evaluation model for language '{self.lang}' into '{file_path}'.")
